-- |
-- Load a configuration file.
-- Syntax (parser is case-sensitive):
-- @
--    features:
--      A
--      B
--      C
--
--    model:
--      A implies B;
--      C and D implies not E;
--      (F and G) or (H and (not I));
-- @
--
module FFJPL.Load where {

import FFJ.Adapter;
import qualified FFJ.Parser;
import FFJ.Parser (PStat);
import FFJPL.FeatureModel;
import FFJPL.Definitions;
import FFJPL.Functions;

import Text.ParserCombinators.Parsec;
import Text.ParserCombinators.Parsec.Expr;
import Text.ParserCombinators.Parsec.Language;
import qualified Text.ParserCombinators.Parsec.Token as PT;

import System.Directory;
import System.FilePath.Posix;

-- |
-- File extension of the files to load
--
fileExtension :: String;
fileExtension = ".ffj";

-- |
-- Load a given configuration file and then load and parse
-- the features specified in the configuration file.
--
load :: FilePath -> IO (CT,PT,([Feature],[Formula]));
load configFile =
    do { config <- readFile configFile;
         -- if parsing fails, throw the parser's error message as
         -- a runtime error.
         let { (feats, fms) = case runParser parseConfig () "" config of
                                  Right r  -> r;
                                  Left err -> error (show err);
             };
         (ct,pt) <- loadFeatures path feats;
         return (ct,pt,("Base":feats,fms));
       }
    where {
      (path, _) = splitFileName configFile
    };

-- |
-- Load all given features (relative to the given path).
--
loadFeatures :: String -> [Feature] -> IO (CT,PT);
loadFeatures path feats = 
    do { pf <- mapM (loadFeature path) feats;
         -- apply 'getFeature' in the order given by 'feats' (i.e., use foldl!)
         return $ foldl (\ctpt (ps,f) -> getFeature ctpt f ps) (initCT,[]) $ concat pf
       };

-- |
-- Load a feature (relative to the given path).
-- Return the name of the feature together with every [PStat]
-- as a convenience for 'loadFeatures' ('getFeature' needs the feature name).
--
loadFeature :: String -> Feature -> IO [([PStat],Feature)];
loadFeature path feat =
    do { contents <- getDirectoryContents fullpath;
         let { toload = [ combine fullpath name 
                        | name <- filter ((==fileExtension).takeExtension) contents ] };
         l <- mapM loadFile toload;
         return [(ps,feat) | ps <- l]
       }
    where {
      fullpath = combine path feat
    };

loadFile :: FilePath -> IO [PStat];
loadFile f = do { putStrLn ("loading '" ++ f ++ "'");
                  contents <- readFile f;
                  return $ (FFJ.Parser.parser . FFJ.Parser.lexer) contents
                };

--
-- Tokenizer
-- We use  haskellStyle  as a basis for the token parser
-- such that identifiers etc. are defined as in Haskell
--
lexer :: PT.TokenParser ();
lexer  = PT.makeTokenParser 
         (haskellStyle
         { reservedNames   = [kwFeatures, kwModel
                             , "and","or","implies","not"]
         , reservedOpNames = [":", "&&", "||", "=>"]
         });
         
kwFeatures, kwModel :: String;
kwFeatures = "features";
kwModel = "model";

reserved   = PT.reserved   lexer;
reservedOp = PT.reservedOp lexer;
parens     = PT.parens     lexer;
semi       = PT.semi       lexer;
identifier = PT.identifier lexer;
-- naturalOrFloat = PT.naturalOrFloat lexer
-- stringLiteral  = PT.stringLiteral  lexer

parseConfig :: Parser ([Feature], [Formula]);
parseConfig = do { fs <- features; ms <- formulas; eof; return (fs,ms) };

features :: Parser [Feature];
features = do { reserved kwFeatures;
                reservedOp ":";
                many identifier
              };

formulas :: Parser [Formula];
formulas = do { reserved kwModel;
                reservedOp ":";
                endBy formula semi;
              };  

formula :: Parser Formula;
formula = buildExpressionParser table operand
        <?> "formula";

--
-- Operator table.
-- "and" bindest strongest, followed by "or", followed by "implies".
-- "and" and "or" are left-associative; "implies" has no associativity.
--
table   = [[op [reserved "and", reservedOp "&&"] (\a b -> And [a,b]) AssocLeft ]
          ,[op [reserved "or", reservedOp "||"] (\a b -> Or [a,b]) AssocLeft]
          ,[op [reserved "implies", reservedOp "=>"] Impl AssocNone]
          ]          
    where {
    op junctors f assoc =
      Infix (do { choice junctors; return f } <?> "operator") assoc;
    };
    
operand :: Parser Formula;
operand = do { reserved "not";     -- negated expression
               f <- operand;
               return $ Not f;
             }
      <|> parens formula           -- expression in parentheses
      <|> featureName;             -- identifier
      
featureName :: Parser Formula;
featureName = do { f <- identifier;
                   return $ FVar f
                 };

}